"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = Channel;
exports.revisionCache = void 0;

var _inherits = _interopRequireDefault(require("inherits"));

var _events = require("events");

var _util = require("./util");

var _uuid = require("uuid");

var _operation = _interopRequireDefault(require("./util/operation"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*eslint no-shadow: 0*/
var UNKNOWN_CV = '?';
var CODE_INVALID_VERSION = 405;
var CODE_DUPLICATE_CHANGE = 409;
var CODE_EMPTY_RESPONSE = 412;
var CODE_INVALID_DIFF = 440;
var operation = {
  MODIFY: _util.change.type.MODIFY,
  REMOVE: _util.change.type.REMOVE
}; // internal methods used as instance methods on a Channel instance

var internal = {};
/**
 * Updates the currently known synced `cv`.
 *
 * @param {String} cv - the change version synced
 * @returns {Promise<String>} the saved `cv`
 */

internal.updateChangeVersion = function (cv) {
  var _this = this;

  return this.store.setChangeVersion(cv).then(function () {
    // A unit test currently relies on this event, otherwise we can remove it
    _this.emit('change-version', cv);

    return cv;
  });
};
/**
 * Creates a change operation for the object of `id` that changes
 * from the date stored in the `ghost` into the data of `object`.
 *
 * Queues the change for syncing.
 *
 * @param {String} id - object id
 * @param {Object} object - object literal of the data that the change should produce
 * @param {Object} ghost - the ghost version used to produce the change object
 */


internal.buildModifyChange = function (id, object, ghost) {
  var payload = _util.change.buildChange(_util.change.type.MODIFY, id, object, ghost),
      empty = true,
      key;

  for (key in payload.v) {
    if (key) {
      empty = false;
      break;
    }
  } // if the change v is an empty object, do not send, notify?


  if (empty) {
    this.emit('unmodified', id, object, ghost);
    return;
  }

  this.localQueue.queue({
    type: 'modify',
    id: id,
    object: object
  });
};
/**
 * Updates the ghost store with the updated ghost data based on the data that
 * comes from applying the patch to the original.
 *
 * @param { string } id - the bucket object key for the object being updated
 * @param { number } version - the version number the ghost is being updated to
 * @param { Object } data - the new data for the ghost for this version
 * @param { Object } original - the original data for the ghost before patch is applied
 * @param { Object } patch - the patch applied to the original to get the resulting data
 * @param { Object } [acknowledged ] - the change sent by this client matching the received network change
 * @returns { Promise<*> } resolves once the ghost has been updated
 */


internal.updateObjectVersion = function (id, version, data, original, patch, acknowledged) {
  var _this2 = this;

  var save = function save() {
    return _this2.store.put(id, version, data);
  };

  if (acknowledged) {
    return save().then(function () {
      internal.updateAcknowledged.call(_this2, acknowledged);
    });
  } // If it's not an ack, it's a change initiated on a different client
  // we need to provide a way for the current client to respond to
  // a potential conflict if it has modifications that have not been synced


  return this.onBeforeNetworkChange(id, data, original, patch).then(function (local) {
    // pull all pending changes in the local queue for the object of id
    var localModifications = _util.change.diff(original, local),
        // rebases the local changes on top of the network patch
    transformed = _util.change.transform(localModifications, patch, original); // remove pending changes


    _this2.localQueue.dequeueChangesFor(id);

    return save().then(function () {
      var update = data; // if the rebase operation results in a modified object
      // generate a new patch and queue it to be sent to simperium

      if (transformed) {
        update = _util.change.apply(transformed, data); // queue up the new change

        _this2.localQueue.queue({
          type: 'modify',
          id: id,
          object: update
        });
      }

      _this2.emit('update', id, update, original, patch, _this2.isIndexing);
    });
  });
};

internal.removeObject = function (id, acknowledged) {
  var notify;

  if (!acknowledged) {
    notify = this.emit.bind(this, 'remove', id);
  } else {
    notify = internal.updateAcknowledged.bind(this, acknowledged);
  }

  return this.store.remove(id).then(notify);
};

internal.updateAcknowledged = function (change) {
  var id = change.id;

  if (this.localQueue.sent[id] === change) {
    this.localQueue.acknowledge(change);
    this.emit('acknowledge', id, change);
  }
};

internal.findAcknowledgedChange = function (change) {
  var possibleChange = this.localQueue.sent[change.id];

  if (possibleChange) {
    if ((change.ccids || []).indexOf(possibleChange.ccid) > -1) {
      return possibleChange;
    }
  }
};

internal.requestObjectVersion = function (id, version) {
  var _this3 = this;

  return new Promise(function (resolve) {
    _this3.once("version.".concat(id, ".").concat(version), function (data) {
      resolve(data);
    });

    _this3.send("e:".concat(id, ".").concat(version));
  });
};

internal.applyChange = function (change, ghost) {
  var _this4 = this;

  var acknowledged = internal.findAcknowledgedChange.call(this, change),
      updateChangeVersion = internal.updateChangeVersion.bind(this, change.cv);
  var error, original, patch, modified; // attempt to apply the change
  // TODO: Handle errors as specified in
  //	 0:c:[{"ccids": ["0435edf4-3f07-4cc6-bf86-f68e6db8779c"], "id": "9e9a9616-8174-42
  // { ccids: [ '0435edf4-3f07-4cc6-bf86-f68e6db8779c' ],
  //	 id: '9e9a9616-8174-425a-a1b0-9ed5410f1edc',
  //	 clientid: 'node-b9776e96-c068-42ae-893a-03f50833bddb',
  //	 error: 400 }

  if (change.error) {
    error = new Error("".concat(change.error, " - Could not apply change to: ").concat(ghost.key));
    error.code = change.error;
    error.change = change;
    error.ghost = ghost;
    internal.handleChangeError.call(this, error, change, acknowledged);
    return;
  }

  if (change.o === operation.MODIFY) {
    if (ghost && ghost.version !== change.sv) {
      internal.requestObjectVersion.call(this, change.id, change.sv).then(function (data) {
        internal.applyChange.call(_this4, change, {
          version: change.sv,
          data: data
        });
      });
      return;
    }

    original = ghost.data;
    patch = change.v;
    modified = _util.change.apply(patch, original);
    return internal.updateObjectVersion.call(this, change.id, change.ev, modified, original, patch, acknowledged).then(updateChangeVersion);
  } else if (change.o === operation.REMOVE) {
    return internal.removeObject.bind(this)(change.id, acknowledged).then(updateChangeVersion);
  }
};

internal.handleChangeError = function (err, change, acknowledged) {
  var _this5 = this;

  switch (err.code) {
    case CODE_INVALID_VERSION:
    case CODE_INVALID_DIFF:
      // Invalid version or diff, send full object back to server
      if (!acknowledged || !acknowledged.d) {
        this.store.get(change.id).then(function (object) {
          _this5.localQueue.queue({
            type: 'full',
            originalChange: acknowledged,
            object: object
          });
        });
      } else {
        this.localQueue.dequeueChangesFor(change.id);
      }

      break;

    case CODE_DUPLICATE_CHANGE:
      // no need to do anything else here - we already got this
      internal.updateAcknowledged.call(this, acknowledged);
      break;

    case CODE_EMPTY_RESPONSE:
      // Change causes no change, just acknowledge it
      internal.updateAcknowledged.call(this, acknowledged);
      break;

    default:
      this.emit('error', err, change);
  }
};

internal.indexingComplete = function () {
  var _this6 = this;

  // Indexing has finished
  this.setIsIndexing(false);
  internal.updateChangeVersion.call(this, this.index_cv).then(function () {
    _this6.localQueue.start();
  });
  this.emit('index', this.index_cv);
  this.index_last_id = null;
  this.index_cv = null;
  this.emit('ready');
};
/**
 * A ghost represents a version of a bucket object as known by Simperium
 *
 * Generally a client will keep the last known ghost stored locally for efficient
 * diffing and patching of Simperium change operations.
 *
 * @typedef {Object} Ghost
 * @property {Number} version - the ghost's version
 * @property {String} key - the simperium bucket object id this ghost is for
 * @property {Object} data - the data for the given ghost version
 */

/**
 * Callback function used by the ghost store to iterate over existing ghosts
 *
 * @callback ghostIterator
 * @param {Ghost} - the current ghost
 */

/**
 * A GhostStore provides the store mechanism for ghost data that the Channel
 * uses to maintain syncing state and producing change operations for
 * Bucket objects.
 *
 * @interface GhostStore
 */

/**
 * Retrieve a Ghost for the given bucket object id
 *
 * @function
 * @name GhostStore#get
 * @param {String} id - bucket object id
 * @returns {Promise<Ghost>} - the ghost for this object
 */

/**
 * Save a ghost in the store.
 *
 * @function
 * @name GhostStore#put
 * @param {String} id - bucket object id
 * @param {Number} version - version of ghost data
 * @param {Object} data - object literal to save as this ghost's data for this version
 * @returns {Promise<Ghost>} - the ghost for this object
 */

/**
 * Delete a Ghost from the store.
 *
 * @function
 * @name GhostStore#remove
 * @param {String} id - bucket object id
 * @returns {Promise<Ghost>} - the ghost for this object
 */

/**
 * Iterate over existing Ghost objects with the given callback.
 *
 * @function
 * @name GhostStore#eachGhost
 * @param {ghostIterator} - function to run against each ghost
 */

/**
 * Get the current change version (cv) that this channel has synced.
 *
 * @function
 * @name GhostStore#getChangeVersion
 * @returns {Promise<String>} - the current change version for the bucket
 */

/**
 * Set the current change version.
 *
 * @function
 * @name GhostStore#setChangeVersion
 * @returns {Promise<Void>} - resolves once the change version is saved
 */

/**
 * Maintains syncing state for a Simperium bucket.
 *
 * A bucket uses a channel to listen for updates that come from simperium while
 * sending updates that are made on the client.
 *
 * The channel can handle incoming simperium commands via `handleMessage`. These
 * messages are stripped of their channel number that separates bucket operations.
 * The `Client` maintains which commands should be routed to which channel.
 *
 * The channel is responsible for creating all change operations and downloading
 * bucket data.
 *
 * @param {String} appid - Simperium app id, used for authenticating
 * @param {String} access_token - Simperium user access token
 * @param {GhostStore} store - data storage for ghost objects
 * @param {String} name - the name of the bucket on Simperium.com
 */


function Channel(appid, access_token, store, name) {
  var _this7 = this;

  // Uses an event emitter to handle different Simperium  commands
  var message = this.message = new _events.EventEmitter();
  this.name = name;
  this.isIndexing = false;
  this.appid = appid;
  this.store = store;
  this.access_token = access_token;
  this.session_id = 'node-' + (0, _uuid.v4)(); // These are the simperium bucket commands the channel knows how to handle

  message.on('auth', this.onAuth.bind(this));
  message.on('i', this.onIndex.bind(this));
  message.on('c', this.onChanges.bind(this));
  message.on('e', this.onVersion.bind(this));
  message.on('cv', this.onChangeVersion.bind(this));
  message.on('o', function () {}); // Maintain a queue of operations that come from simperium commands
  // so that the can be applied to the ghost data.

  this.networkQueue = new NetworkQueue(); // Maintain a queue of operations that originate from this client
  // to track their status.

  this.localQueue = new LocalQueue(this.store); // When a local queue has indicatie that it should send a change operation
  // emit a simperium command. The Client instance will know how to route that
  // command correctly to simperium

  this.localQueue.on('send', function (data) {
    _this7.emit('send', "c:".concat(JSON.stringify(data)));
  }); // Handle change errors caused by changes originating from this client

  this.localQueue.on('error', internal.handleChangeError.bind(this));
}

(0, _inherits.default)(Channel, _events.EventEmitter);
/**
 * Called by a bucket when a bucket object has been updated.
 *
 * The channel uses this method to initiate change operations when objects are updated.
 *
 * It also uses this method during indexing to track which objects have been successfully
 * downloaded.
 *
 * @param {BucketObject} object - the bucket object
 * @param {Boolean} [sync=true] - if the object should be synced
 * @returns {Promise<BucketObject>} the bucket object
 */

Channel.prototype.update = function (object) {
  var _this8 = this;

  var sync = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
  this.onBucketUpdate(object.id);

  if (sync === true) {
    return this.store.get(object.id).then(function (ghost) {
      return internal.buildModifyChange.call(_this8, object.id, object.data, ghost);
    }).then(function () {
      return object;
    });
  }

  return Promise.resolve(object);
};
/**
 * Tracks indexing state and emits `indexingStateChange`
 *
 * @private
 * @param {Boolean} isIndexing - updates indexing state to this value
 */


Channel.prototype.setIsIndexing = function (isIndexing) {
  this.isIndexing = isIndexing;
  this.emit('indexingStateChange', this.isIndexing);
};
/**
 * Removes an object from Simperium. Called by a bucket when an object is deleted.
 *
 * @param {String} id - the id of the object to remove
 */


Channel.prototype.remove = function (id) {
  var _this9 = this;

  this.store.get(id).then(function () {
    return _this9.localQueue.queue({
      type: 'remove',
      id: id
    });
  });
};
/**
 * Retrieves revisions for a given object from Simperium.
 *
 * @typedef {Object} BucketObjectRevision
 * @property {String} id - bucket object id
 * @property {Number} version - revision version
 * @property {Object} data - object literal data at given version
 *
 * @param {String} id - the bucket object id
 * @returns {Promise<Array<BucketObjectRevision>>} list of known object versions
 */


Channel.prototype.getRevisions = function (id) {
  var _this10 = this;

  return new Promise(function (resolve, reject) {
    collectionRevisions(_this10, id, function (error, revisions) {
      if (error) {
        reject(error);
        return;
      }

      resolve(revisions);
    });
  });
};
/**
 * Checks if there are unsynced changes.
 *
 * @returns {Promise<Boolean>} true if there are still changes to sync
 */


Channel.prototype.hasLocalChanges = function () {
  return Promise.resolve(Object.keys(this.localQueue.queues).length > 0);
};
/**
 * Retrieves the currently stored version number for a given object
 *
 * @param {String} id - object id to get the version for
 * @returns {Promise<Number>} version number for the object
 */


Channel.prototype.getVersion = function (id) {
  return this.store.get(id).then(function (ghost) {
    if (ghost && ghost.version) {
      return ghost.version;
    }

    return 0;
  });
};
/**
 * Subscription interface for network changes.
 * @param { NetworkChangeResolver } changeResolver - function used to subscribe to network changes modifying a
 *                                    bucket object
 */


Channel.prototype.beforeNetworkChange = function (changeResolver) {
  this.changeResolver = changeResolver;
};

Channel.prototype.onBeforeNetworkChange = function (id, data, base, patch) {
  if (this.changeResolver) {
    return Promise.resolve(this.changeResolver(id, data, base, patch));
  }

  return Promise.resolve();
};
/**
 * Receives incoming messages from Simperium
 *
 * Called by a client that strips the channel number prefix before
 * seding to a specific channel.
 *
 * @param {String} data - the message from Simperium
 */


Channel.prototype.handleMessage = function (data) {
  var message = (0, _util.parseMessage)(data);
  this.message.emit(message.command, message.data);
};
/**
 * Used to send a message from this channel to Simperium
 * The client listens for `send` events and correctly sends them to Simperium
 *
 * @emits Channel#send
 * @private
 * @param {String} data - the message to send
 */


Channel.prototype.send = function (data) {
  /**
   * Send event
   *
   * @event Channel#send
   * @type {String} - the message to send to Simperium
   */
  this.emit('send', data);
};
/**
 * Restores a buckets data to what is currently stored in the ghost data.
 */


Channel.prototype.reload = function () {
  var _this11 = this;

  this.store.eachGhost(function (ghost) {
    _this11.emit('update', ghost.key, ghost.data);
  });
};
/**
 * Called after a bucket updates an object.
 *
 * Wile indexing keeps track of which objects have been retrieved.
 *
 * @param {String} id - object that was updated
 */


Channel.prototype.onBucketUpdate = function (id) {
  if (!this.isIndexing) {
    return;
  }

  if (this.index_last_id == null || this.index_cv == null) {
    return;
  } else if (this.index_last_id === id) {
    internal.indexingComplete.call(this);
  }
};

Channel.prototype.onAuth = function (data) {
  var _this12 = this;

  var auth;
  var init;

  try {
    auth = JSON.parse(data);
    this.emit('unauthorized', auth);
    return;
  } catch (error) {
    // request cv and then send method
    this.once('ready', function () {
      _this12.localQueue.resendSentChanges();
    });

    init = function init(cv) {
      if (cv) {
        _this12.localQueue.start();

        _this12.sendChangeVersionRequest(cv);
      } else {
        _this12.startIndexing();
      }
    };

    this.store.getChangeVersion().then(init);
    return;
  }
};
/**
 * Re-downloads all Simperium bucket data
 */


Channel.prototype.startIndexing = function () {
  this.localQueue.pause();
  this.setIsIndexing(true);
  this.sendIndexRequest();
};
/**
 * Called when a channel's socket has been connected
 */


Channel.prototype.onConnect = function () {
  var init = {
    name: this.name,
    clientid: this.session_id,
    api: '1.1',
    token: this.access_token,
    app_id: this.appid,
    library: 'node-simperium',
    version: '0.0.1'
  };
  this.send("init:".concat(JSON.stringify(init)));
};

Channel.prototype.onIndex = function (data) {
  var page = JSON.parse(data),
      objects = page.index,
      mark = page.mark,
      cv = page.current,
      update = internal.updateObjectVersion.bind(this);
  var objectId;
  objects.forEach(function (object) {
    objectId = object.id;
    update(object.id, object.v, object.d);
  });

  if (!mark) {
    if (objectId) {
      this.index_last_id = objectId;
    }

    if (!this.index_last_id) {
      internal.indexingComplete.call(this);
    }

    this.index_cv = cv;
  } else {
    this.sendIndexRequest(mark);
  }
};

Channel.prototype.sendIndexRequest = function (mark) {
  this.send("i:1:".concat(mark ? mark : '', "::10"));
};

Channel.prototype.sendChangeVersionRequest = function (cv) {
  this.send("cv:".concat(cv));
};

Channel.prototype.onChanges = function (changes) {
  var _this13 = this;

  JSON.parse(changes).forEach(function (change) {
    // pull out the object from the store and apply the change delta
    _this13.networkQueue.queueFor(change.id).add(function (done) {
      return _this13.store.get(change.id).then(function (ghost) {
        return internal.applyChange.call(_this13, change, ghost);
      }).then(done, done);
    });
  }); // emit ready after all server changes have been applied

  this.emit('ready');
};

Channel.prototype.onChangeVersion = function (data) {
  var _this14 = this;

  if (data === UNKNOWN_CV) {
    this.store.setChangeVersion(null).then(function () {
      return _this14.startIndexing();
    });
  }
};

Channel.prototype.onVersion = function (data) {
  // invalid version, give up without emitting
  if (data.slice(-2) === '\n?') {
    return;
  }

  var ghost = (0, _util.parseVersionMessage)(data);
  this.emit('version', ghost.id, ghost.version, ghost.data);
  this.emit('version.' + ghost.id, ghost.id, ghost.version, ghost.data);
  this.emit('version.' + ghost.id + '.' + ghost.version, ghost.data);
};

function NetworkQueue() {
  this.queues = {};
}

NetworkQueue.prototype.queueFor = function (id) {
  var queues = this.queues,
      queue = queues[id];

  if (!queue) {
    queue = new Queue();
    queue.on('finish', function () {
      delete queues[id];
    });
    queues[id] = queue;
  }

  return queue;
};

function Queue() {
  this.queue = [];
  this.running = false;
}

(0, _inherits.default)(Queue, _events.EventEmitter); // Add a function at the end of the queue

Queue.prototype.add = function (fn) {
  this.queue.push(fn);
  this.start();
  return this;
};

Queue.prototype.start = function () {
  if (this.running) return;
  this.running = true;
  this.emit('start');
  setImmediate(this.run.bind(this));
};

Queue.prototype.run = function () {
  var fn;
  this.running = true;

  if (this.queue.length === 0) {
    this.running = false;
    this.emit('finish');
    return;
  }

  fn = this.queue.shift();
  fn(this.run.bind(this));
};

function LocalQueue(store) {
  this.store = store;
  this.sent = {};
  this.queues = {};
  this.ready = false;
}

(0, _inherits.default)(LocalQueue, _events.EventEmitter);

LocalQueue.prototype.start = function () {
  var queueId;
  this.ready = true;

  for (queueId in this.queues) {
    this.processQueue(queueId);
  }
};

LocalQueue.prototype.pause = function () {
  this.ready = false;
};

LocalQueue.prototype.acknowledge = function (change) {
  if (this.sent[change.id] === change) {
    delete this.sent[change.id];
  }

  this.processQueue(change.id);
};

LocalQueue.prototype.queue = function (change) {
  var queue = this.queues[change.id];

  if (!queue) {
    queue = [];
    this.queues[change.id] = queue;
  }

  queue.push(change);
  this.emit('queued', change.id, change, queue);
  if (!this.ready) return;
  this.processQueue(change.id);
};

LocalQueue.prototype.dequeueChangesFor = function (id) {
  var changes = [],
      sent = this.sent[id],
      queue = this.queues[id];

  if (sent) {
    changes.push(sent);
  }

  if (queue) {
    delete this.queues[id];
    changes = changes.concat(queue);
  }

  return changes;
};

LocalQueue.prototype.processQueue = function (id) {
  var _this15 = this;

  var queue = this.queues[id]; // there is no queue, don't do anything

  if (!queue) return; // queue is empty, delete it from memory

  if (queue.length === 0) {
    delete this.queues[id];
    return;
  } // waiting for a previous sent change to get acknowledged


  if (this.sent[id]) {
    this.emit('wait', id);
    return;
  }

  this.store.get(id).then(function (ghost) {
    var changes = _this15.queues[id]; // a change was sent before we could compress and send

    if (_this15.sent[id]) {
      _this15.emit('wait', id);

      return;
    }

    var sending = changes.reduce(function (chosen, next) {
      return 'remove' === chosen.type ? chosen : next;
    });
    var change = (0, _operation.default)(sending, ghost);
    _this15.queues[id] = [];

    if (_util.change.isEmptyChange(change)) {
      return;
    }

    _this15.sent[id] = change;

    _this15.emit('send', change);
  });
};

LocalQueue.prototype.resendSentChanges = function () {
  for (var ccid in this.sent) {
    this.emit('send', this.sent[ccid]);
  }
};
/**
 * Since revision data is basically immutable we can prevent the
 * need to refetch it after it has been loaded once.
 *
 * E.g. key could be `${ entityId }.${ versionNumber }`
 *
 * @type {Map<String,Object>} stores specific revisions as a cache
 */


var revisionCache = new Map();
/**
 * Attempts to fetch an entity's revisions
 *
 * By default, a bucket stores two kinds of history:
 * 	- revisions: the most-recent changes to an entity (60 of these)
 * 	- archive: a "snapshot" of every ten revisions (100 of these)
 *
 * Together the revisions and archive span changes over the
 * 1,060 most-recent changes to an entity, but of course once
 * we hit the archive we lose save granularity.
 *
 * Individual buckets can override the defaults as well and also
 * completely eliminate them.
 *
 * We don't have a listing of which revisions exist for a given entity.
 *
 * @param {Object} channel used to send messages to the Simperium server
 * @param {String} id entity id for which to fetch revisions
 * @param {Function} callback called on error or when finished
 */

exports.revisionCache = revisionCache;

function collectionRevisions(channel, id, callback) {
  /** @type {Number} ms delay arbitrarily chosen to give up on fetch */
  var TIMEOUT = 200;
  /** @type {Set} tracks requested revisions */

  var requestedVersions = new Set();
  /** @type {Array<Object>} contains the revisions and associated data */

  var versions = [];
  /** @type {Number} remembers newest version of an entity */

  var latestVersion;
  /** @type {Number} handle for "start finishing" timeout */

  var timeout;
  /**
   * Receive a version update from the server and
   * dispatch the next fetch or finish the fetching
   *
   * @param {String} id entity id
   * @param {Number} version version of returned entity
   * @param {Object} data value of entity at revision
   */

  function onVersion(id, version, data) {
    revisionCache.set("".concat(id, ".").concat(version), data);
    versions.push({
      id: id,
      version: version,
      data: data
    }); // if we have every possible revision already, finish it!
    // this bypasses any mandatory delay

    if (versions.length === latestVersion) {
      finish();
      return;
    }

    fetchNextVersion(version); // defer the final response to the application

    clearTimeout(timeout);
    timeout = setTimeout(finish, TIMEOUT);
  }
  /**
   * Stop listening for versions and stop fetching them
   * and pass accumulated data back to application
   */


  function finish() {
    clearTimeout(timeout);
    channel.removeListener("version.".concat(id), onVersion); // sort newest first

    callback(null, versions.sort(function (a, b) {
      return b.version - a.version;
    }));
  }
  /**
   * Find the next version which isn't around and issue
   * a fetch if possible
   *
   * @param {Number} prevVersion starting point for finding next version
   */


  function fetchNextVersion(prevVersion) {
    var version = prevVersion; // find the next version to request
    // some could have come back already
    // or been requested already

    while (version > 0 && requestedVersions.has(version)) {
      version -= 1;
    } // we have them all


    if (!version) {
      return;
    }

    requestedVersions.add(version); // fetch from server or local cache

    if (revisionCache.has("".concat(id, ".").concat(version))) {
      onVersion(id, version, revisionCache.get("".concat(id, ".").concat(version)));
    } else {
      channel.send("e:".concat(id, ".").concat(version));
    }
  } // start listening for the responses


  channel.on("version.".concat(id), onVersion); // request the first revision and start the sequence
  // pre-emptively fetch as many as could exist by default

  channel.store.get(id).then(function (_ref) {
    var version = _ref.version;
    latestVersion = version; // grab latest change revisions

    for (var i = 0; i < 60 && version - i > 0; i++) {
      fetchNextVersion(version - i);
    } // grab archive revisions
    // these are like 1, 11, 21, 31, …, 41, normal revisions [42, 43, 44, 45, …]


    var firstArchive = Math.round((version - 60) / 10) * 10 + 1; // 127 -> 67 -> 6 -> 60 -> 61

    for (var _i = 0; _i < 100 && firstArchive - 10 * _i > 0; _i++) {
      fetchNextVersion(firstArchive - 10 * _i);
    }
  }, callback); // and set an initial timeout for failed connections

  timeout = setTimeout(finish, TIMEOUT * 4);
}