"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = Client;
exports.Heartbeat = Heartbeat;
exports.ReconnectionTimer = ReconnectionTimer;
Object.defineProperty(exports, "Bucket", {
  enumerable: true,
  get: function get() {
    return _bucket.default;
  }
});
Object.defineProperty(exports, "Channel", {
  enumerable: true,
  get: function get() {
    return _channel.default;
  }
});

var _inherits = _interopRequireDefault(require("inherits"));

var _events = require("events");

var _bucket = _interopRequireDefault(require("./bucket"));

var _channel = _interopRequireDefault(require("./channel"));

var _default = _interopRequireDefault(require("./ghost/default"));

var _default2 = _interopRequireDefault(require("./storage/default"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @function
 * @name bucketStoreProvider
 * @param {Bucket} - the bucket to create a store instance for
 * @returns {BucketStore} - the bucket store instance to be used by the bucket
 */

/**
 * @function
 * @name ghostStoreProvider
 * @param {Bucket} - the bucket to create a store instance for
 * @returns {GhostStore} - the ghost store instance to be used by the bucket
 */

/**
 * @function
 * @name websocketClientProvider
 * @param {String} - The url to open the socket connection with
 * @returns {Object} a WebSocket client

/**
 * A Client is the main interface to Simperium.
 *
 * @param {String} appId - Simperium application id
 * @param {String} accessToken - User access token
 * @param {Object} options - configuration options for the client
 * @param {ghostStoreProvider} [options.ghostStoreProvider=defaultGhostStoreProvider]
 *            - factory function for creating ghost store instances
 * @param {bucketStoreProvider} [options.objectStoreProvider=defaultObjectStoreProvider]
 *            - factory function for creating object store instances
 * @param {number} [options.heartbeatInterval=4] - heartbeat interval for maintaining connection status with Simperium.com
 * @param {websocketClientProvider} [options.websocketClientProvider] - WebSocket transport, if not provided tries to use window.WebSocket
 */
function Client(appId, accessToken, options) {
  options = options || {};
  options.ghostStoreProvider = options.ghostStoreProvider || _default.default;
  options.objectStoreProvider = options.objectStoreProvider || _default2.default;
  options.hearbeatInterval = options.heartbeatInterval || 4;
  options.websocketClientProvider = options.websocketClientProvider || defaultWebsocketClientProvider;
  this.accessToken = accessToken;
  this.open = false;
  this.options = options;
  this.heartbeat = new Heartbeat(options.hearbeatInterval, this.sendHeartbeat.bind(this));
  this.heartbeat.on('timeout', this.onConnectionTimeout.bind(this));
  this.reconnectionTimer = new ReconnectionTimer(function (attempt) {
    var time = (attempt >= 3 ? (attempt - 3) * 3000 : 0) + 3000;
    return time;
  }, this.onReconnect.bind(this));
  this.appId = appId;
  options.url = options.url || "wss://api.simperium.com/sock/1/".concat(this.appId, "/websocket");
  this.reconnect = true;
  this.on('message:h', this.onHeartbeat.bind(this));
  this.buckets = [];
  this.connect();
}

(0, _inherits.default)(Client, _events.EventEmitter);
/**
 * Set up a bucket with the given name for interacting with Simperium.
 *
 * @param {String} name - the bucket name on simperium
 * @returns {Bucket} a bucket instance configured for syncing
 */

Client.prototype.bucket = function (name) {
  var channelId = this.buckets.length,
      bucket = new _bucket.default(name, this.options.objectStoreProvider),
      channel = new _channel.default(this.appId, this.accessToken, this.options.ghostStoreProvider(bucket), name),
      send = this.sendChannelMessage.bind(this, channelId),
      receive = channel.handleMessage.bind(channel);
  bucket.setChannel(channel);
  this.buckets.push(bucket);
  channel.on('unauthorized', this.onUnauthorized.bind(this));
  channel.on('send', send);
  this.on('connect', channel.onConnect.bind(channel));
  this.on("channel:".concat(channelId), receive);
  this.on('access-token', function (token) {
    channel.access_token = token;
  });
  if (this.open) channel.onConnect();
  return bucket;
};

Client.prototype.onHeartbeat = function (message) {
  var counter = parseInt(message);
  this.heartbeat.tick(counter);
};

Client.prototype.onConnect = function () {
  this.open = true;
  this.heartbeat.start();
  this.reconnectionTimer.reset();
  this.emit('connect');
};

Client.prototype.onReconnect = function (attempt) {
  this.emit('reconnect', attempt);
  this.connect();
};

Client.prototype.onConnectionTimeout = function () {
  this.disconnect();
};

Client.prototype.onConnectionFailed = function () {
  this.emit('disconnect');
  if (this.reconnect) this.reconnectionTimer.start();
};

Client.prototype.onMessage = function (message) {
  this.parseMessage(message);
  this.heartbeat.tick();
};

Client.prototype.onUnauthorized = function (details) {
  this.reconnect = false;
  this.emit('unauthorized', details);
};

Client.prototype.parseMessage = function (event) {
  var data = event.data,
      marker = data.indexOf(':'),
      prefix = data.slice(0, marker),
      channelId = parseInt(prefix),
      channelMessage = data.slice(marker + 1);
  this.emit('message', data);

  if (isNaN(channelId)) {
    this.emit("message:".concat(prefix), channelMessage);
  } else {
    this.emit("channel:".concat(channelId), channelMessage);
  }
};

Client.prototype.sendHeartbeat = function (count) {
  this.send("h:".concat(count));
};

Client.prototype.send = function (data) {
  this.emit('send', data);

  try {
    this.socket.send(data);
  } catch (e) {
    // failed to send, probably not connected
    this.emit('error', e);
  }
};

Client.prototype.sendChannelMessage = function (id, message) {
  this.send("".concat(id, ":").concat(message));
};

Client.prototype.connect = function () {
  if (this.open) {
    this.disconnect();
  }

  this.reconnect = true;
  this.socket = this.options.websocketClientProvider(this.options.url);
  this.socket.onopen = this.onConnect.bind(this);
  this.socket.onmessage = this.onMessage.bind(this);
  this.socket.onclose = this.onConnectionFailed.bind(this);
};

Client.prototype.disconnect = function () {
  if (this.open) {
    this.socket.close();
  } else {
    this.onClose();
  }
};

Client.prototype.end = function () {
  this.reconnect = false;
  this.reconnectionTimer.stop();
  this.heartbeat.stop();
  this.disconnect();
};

Client.prototype.onClose = function () {
  this.connection = null;
  this.heartbeat.stop();
  if (this.reconnect !== false) this.reconnectionTimer.start();
  this.emit('close');
};

Client.prototype.setAccessToken = function (token) {
  this.accessToken = token;
  this.emit('access-token', token);
  this.connect();
};

function Heartbeat(seconds, onBeat) {
  this.count = 0;
  this.seconds = seconds;

  _events.EventEmitter.call(this);

  if (onBeat) this.on('beat', onBeat);
}

(0, _inherits.default)(Heartbeat, _events.EventEmitter);

Heartbeat.prototype.onBeat = function () {
  this.count++;
  this.timeout = setTimeout(this.onTimeout.bind(this), this.seconds * 1000 * 2);
  this.emit('beat', this.count);
};

Heartbeat.prototype.onTimeout = function () {
  this.emit('timeout');
  this.stop();
};

Heartbeat.prototype.tick = function (count) {
  if (count > 0 && typeof count === 'number') {
    this.count = count;
  }

  this.start();
};

Heartbeat.prototype.start = function () {
  this.stop();
  clearTimeout(this.timer);
  this.timer = setTimeout(this.onBeat.bind(this), this.seconds * 1000);
};

Heartbeat.prototype.stop = function () {
  clearTimeout(this.timer);
  clearTimeout(this.timeout);
};

function ReconnectionTimer(interval, onTripped) {
  _events.EventEmitter.call(this);

  this.started = false;

  this.interval = interval || function () {
    return 1000;
  };

  if (onTripped) this.on('tripped', onTripped);
  this.reset();
}

(0, _inherits.default)(ReconnectionTimer, _events.EventEmitter);

ReconnectionTimer.prototype.onInterval = function () {
  this.emit('tripped', this.attempt);
  this.attempt++;
};

ReconnectionTimer.prototype.start = function () {
  this.started = true;
  this.timer = setTimeout(this.onInterval.bind(this), this.interval(this.attempt));
};

ReconnectionTimer.prototype.restart = function () {
  this.reset();
  this.start();
};

ReconnectionTimer.prototype.reset = ReconnectionTimer.prototype.stop = function () {
  this.attempt = 0;
  this.started = false;
  clearTimeout(this.timer);
};

function defaultWebsocketClientProvider(url) {
  var WebSocketClient;

  if (typeof window !== 'undefined' && window.WebSocket) {
    WebSocketClient = window.WebSocket;
  } else {
    WebSocketClient = require('websocket').w3cwebsocket;
  }

  return new WebSocketClient(url);
} // Simplenote expects this.


Client.Bucket = _bucket.default;